/*
Very crude and simple todo manager for the command line.
written in about an hour as I was teaching myself Go. So don't expext beautiful
code. It just works in a simple and crude way.

Todo items are saved in json format so they can be imported or read by anything
that can read plain text and convert json. An iOS app might follow.

By default, the todo file is saved in your home directory, ~/.todo/todo.json
however, there is an option that allows you to enter a different location. This
option (-c) will let you save the file on a cloud-based directory on your computer,
like Dropbox or Google Drive. The new path is saved to ~/.todo/todo.config

WARNING
Use at your own risk. I gave you the code, don't come knocking if something
didn't work or was deleted. I coded this for me, and it works. I don't care if
it doesn't for you.

Version 0.1: initial version.
Version 0.2: added display of the todo file location. Added the words TODAY and
IMPORTANT to the display of items to make them pop. Added sorting of list by
priority so they items for TODAY show first, followed by IMPORTANT.
Version 0.3: fixed out of bound errors.
Version 0.4: added a shortcut to quickly add an important task with ** and a
priority 0 with TODAY
Version 0.5: added option to display without colors
*/
package main

import (
	"bufio"
	"encoding/json"
	"flag"
	"fmt"
	"io/ioutil"
	"os"
	"os/user"
	"sort"
	"strconv"
	"strings"
)

// main structure of the todo item
type Todo struct {
	Item     string
	Priority int
}

// app version
const VERSION = "0.5.1"

// colors for printf
const KRED = "\x1B[31m" // color red
const KGRN = "\x1B[92m" // color green
const KNRM = "\x1B[0m"  // reset color

// default names for files and directories
const TODOF = "todo.json"
const CONFIGF = "todo.config"
const TODODIR = "/.todo/"

const ACTION_MOD_TEXT = 0
const ACTION_MOD_PRIORITY = 1

// keep the path to the todo file, e.g ~/.todo/todo.json
var todofile string

// configuration file ~/.todo/todo.config containing the todo file path
// store here where you want to the todo.json file to live
// by default is ~/.todo/todo.json but it can be stored on any cloud or
// file sync service directory like ~/Dropbox or /Volume/..../GoogleDrive
var configfile string

// flag for removing colorized output
var removeColor bool

// show app version
func showVersion() {
	fmt.Println("Version", VERSION)
}

// show command line help
func showHelp() {
	fmt.Println("Usage:")
	fmt.Println("\t todo NOARGS: display current todo list")
	fmt.Println("\t todo [-p PRIORITY] STRING: add STRING to the todo list.")
	fmt.Println("\t\t PRIORITY 0 = item for NOW, 1 = important item to do next.")
	fmt.Println("\t\t Use \"**\" as a shortcut to create an important task, and")
	fmt.Println("\t\t \"TODAY\" a task for today.")
	fmt.Println("\t\t eg. ./todo \"** this will be a priority 1 item\"")
	fmt.Println("\t\t If no priority passed, assume normal priority (p = 2).")
	fmt.Println("\t todo ITENUMBER: set ITENUMBER to done, and remove from todo list")
	fmt.Println("\t todo -m ITENUMBER \"new string\": replace ITENUMBER with new string")
	fmt.Println("\t todo -c TODOFILEPATH: set default todo file path to TODOFILEPATH")
	fmt.Println("\t todo -a ITENUMBER: set ITENUMBER to priority 0 (needed now)")
	fmt.Println("\t todo -i ITENUMBER: set ITENUMBER to priority 1 (important)")
	fmt.Println("\t todo -n ITENUMBER: set ITENUMBER to priority 2 (normal)")
	fmt.Println("\t todo -t: display only the tasks marked for today")
	fmt.Println("\t todo -d: display only the tasks marked as important")
	fmt.Println("\t todo -r: disable colorized output")
	fmt.Println("\t todo -f: display location of the todo file")
	fmt.Println("\t todo -h: display this help")
	fmt.Println("\t todo -v: display version")
}

func showTodoFilePath() {
	fmt.Println(todofile)
}

// fileExists checks if a file exists
func fileExists(filename string) bool {
	_, err := os.Stat(filename)
	if os.IsNotExist(err) {
		return false
	}
	return true
}

// get home dir for the user
func getHomeDir() string {
	usr, err := user.Current()
	if err != nil {
		panic(err)
	}
	return usr.HomeDir
}

// check if todo dir exists and if not create it
func createHomeDirIfNotExist(dir string) {
	if _, err := os.Stat(dir); os.IsNotExist(err) {
		//create the local directory
		err = os.MkdirAll(dir, 0755)
		if err != nil {
			panic(err)
		}
	}
}

// create an empty file in the path
func touchFile(fullfilename string) error {
	file, err := os.OpenFile(fullfilename, os.O_RDONLY|os.O_CREATE, 0644)
	if err != nil {
		return err
	}
	return file.Close()
}

// Crude path building. Needs to be improved
// Create the path if it doesn't exists and set the todo file, config file...
func setupLocal() {
	var home string
	home = getHomeDir() + TODODIR
	todofile = home + TODOF
	configfile = home + CONFIGF
	// setup local directory
	createHomeDirIfNotExist(home)

	// setup empty todo file if it doesn't exist
	if !fileExists(todofile) {
		fmt.Println("Creating default todo file in " + todofile + " ...")
		touchFile(todofile)
	}
}

// if user passes a remote place to store the todo list, create the fie there
func setupRemote(path string) {
	var fullremotepath string
	fmt.Println("Creating remote todo file in " + path + " ...")
	createHomeDirIfNotExist(path)
	fullremotepath = path + TODOF
	touchFile(fullremotepath)
	saveConfigFile(fullremotepath)
}

//save path entered by the user to todo.config
func saveConfigFile(line string) {
	f, err := os.Create(configfile)

	if err != nil {
		fmt.Println("Could not open config file: " + configfile)
		os.Exit(1)
	}
	defer f.Close()

	if _, err := f.WriteString(line); err != nil {
		fmt.Println(err)
	}
}

// read config file for external locaton of the TODO file, if it exists
func readConfigFile(configfilepath string) {
	if !fileExists(configfilepath) {
		return
	}

	f, err := os.Open(configfilepath)
	if err != nil {
		fmt.Println("Could not open config file: " + configfilepath)
		os.Exit(1)
	}
	defer f.Close()

	scanner := bufio.NewScanner(f)
	scanner.Scan()
	todofile = scanner.Text()
}

// check if we had an error
func check(e error) {
	if e != nil {
		fmt.Println("Error:", e)
		os.Exit(1)
	}
}

// write current todo to json file
func writeTodo(filename string, todolist []Todo) {
	file, _ := json.MarshalIndent(todolist, "", " ")
	err := ioutil.WriteFile(filename, file, 0644)

	// if err get out
	check(err)
}

// read todo from json file
// return list
func readTodo(filename string) []Todo {

	// open file
	jsonFile, err := os.Open(filename)

	// if err get out
	check(err)

	defer jsonFile.Close()

	// host the current todo list from the file here
	var todolist []Todo

	// read our opened jsonFile as a byte array.
	byteValue, _ := ioutil.ReadAll(jsonFile)

	// unmarshal the byteArray which contains the todo list
	json.Unmarshal(byteValue, &todolist)

	return todolist
}

// show contents of the todo list
func showTodoList(todolist []Todo) {

	if len(todolist) <= 0 {
		fmt.Println("List is empty, go get a coffee.")
		return
	}

	// Color the output based on the "Priority" value:
	// items that need to be done right now (Priority = 0) = Green
	// items that are important and to be done next (Priority = 1) = Red
	// other items (Priority = 2) = Normal
	for i := 0; i < len(todolist); i++ {
		if removeColor {
			switch todolist[i].Priority {
			case 0:
				fmt.Printf("%d. TODAY: %s \n", i, todolist[i].Item)
			case 1:
				fmt.Printf("%d. IMPORTANT: %s \n", i, todolist[i].Item)
			default:
				fmt.Printf("%d. %s \n", i, todolist[i].Item)
			}
		} else {
			switch todolist[i].Priority {
			case 0:
				fmt.Printf("%s%d. TODAY: %s %s\n", KGRN, i, todolist[i].Item, KNRM)
			case 1:
				fmt.Printf("%s%d. IMPORTANT: %s %s\n", KRED, i, todolist[i].Item, KNRM)
			default:
				fmt.Printf("%d. %s \n", i, todolist[i].Item)
			}
		}
	}
}

// show items needed to be done now, priority 0
func showToday(todolist []Todo) {
	for i := 0; i < len(todolist); i++ {
		if todolist[i].Priority == 0 {
			if removeColor {
				fmt.Printf("%d. TODAY: %s \n", i, todolist[i].Item)
			} else {
				fmt.Printf("%s%d. TODAY: %s %s\n", KGRN, i, todolist[i].Item, KNRM)
			}
		}
	}
}

// show items marked as important, priority 1
func showImportant(todolist []Todo) {
	for i := 0; i < len(todolist); i++ {
		if todolist[i].Priority == 1 {
			if removeColor {
				fmt.Printf("%d. IMPORTANT: %s \n", i, todolist[i].Item)
			} else {
				fmt.Printf("%s%d. IMPORTANT: %s %s\n", KRED, i, todolist[i].Item, KNRM)
			}
		}
	}
}

// add item to list
// return updated list
func addItem(todolist []Todo, item string, priority int) []Todo {
	todolist = append(todolist, Todo{Item: item, Priority: priority})
	return todolist
}

// remove item i from the list in todolist
// return updated list of todos
func removeItem(todolist []Todo, i int) []Todo {
	if i > len(todolist)-1 {
		fmt.Println("Error: not such item: ", i)
		os.Exit(1)
	}

	return append(todolist[:i], todolist[i+1:]...)
}

// modify a todo item
// i is the item to be modified
// newitem is the new string to be added
// priority is the new priority to be updated. Pass -1 if not updating this
func modifyItem(todolist []Todo, i int, newitem string, priority int, action int) []Todo {
	if i > len(todolist)-1 {
		fmt.Println("Error: not such item: ", i)
		os.Exit(1)
	}

	// if we are modifying the text and priority
	if action == ACTION_MOD_TEXT {
		todolist[i].Item = newitem
		//if 0, 1, or 2 update also the priority
		if priority > -1 {
			todolist[i].Priority = priority
		}
	} else {
		// if we are modifying only the priority
		todolist[i].Priority = priority
	}
	return todolist
}

// sort the todo list by priority so they show first on the the ouput
func sortTodoList(todolist []Todo) []Todo {
	sort.Slice(todolist, func(i, j int) bool {
		return todolist[i].Priority < todolist[j].Priority
	})

	return todolist
}

// finish action by writing the todo list and displaying it
func finish(todolist []Todo) {

	// sort the list
	sortTodoList(todolist)

	// write current list to file
	writeTodo(todofile, todolist)

	// show the current list
	showTodoList(todolist)
}

// main function
func main() {
	var todolist []Todo
	var priority, todonow, todoimp, normal int
	var item int
	var displayVersion, today, important, filePathLocation bool
	var newTodoFilePath string

	// overwrite the printdefaults
	flag.Usage = showHelp

	// get the priority set, if any. default to 2
	// priorities:
	// 0 = item to be done NOW / TODAY (displayed in green)
	// 1 = important item that needs to be done next (displayed in red)
	// 2 = everything else (displayed in normal color)
	flag.IntVar(&priority, "p", 2, "Priority for the todo item")
	flag.IntVar(&todonow, "a", -1, "set item priority to now")
	flag.IntVar(&todoimp, "i", -1, "set item priority to important")
	flag.IntVar(&normal, "n", -1, "set item priority to normal")

	// select item to replace the string
	flag.IntVar(&item, "m", -1, "Item to replace the string with")

	// show version
	flag.BoolVar(&displayVersion, "v", false, "Show version")
	flag.BoolVar(&today, "t", false, "Show today")
	flag.BoolVar(&important, "d", false, "Show important")
	flag.BoolVar(&filePathLocation, "f", false, "Show important")
	flag.BoolVar(&removeColor, "r", false, "Disable colorized output")

	// pass the location to save the todo.json file to
	flag.StringVar(&newTodoFilePath, "c", "", "Save todo file in this location")

	flag.Parse()

	if displayVersion {
		showVersion()
		return
	}

	// output the location of the todo file
	if filePathLocation {
		showTodoFilePath()
		return
	}

	// check the arguments passed
	arguments := flag.Args()

	// read from file
	todolist = readTodo(todofile)

	// show only the items for today
	if today {
		showToday(todolist)
		return
	}

	// show only the items set as important
	if important {
		showImportant(todolist)
		return
	}

	// check if we need to modify an item
	if item > -1 {
		if flag.Arg(0) == "" {
			fmt.Println("Error: no new string passed to -m. Type todo -h for help")
			return
		}

		todolist = modifyItem(todolist, item, flag.Arg(0), -1, ACTION_MOD_TEXT)
		finish(todolist)
		return
	}

	// set item to NOW
	if todonow > -1 {
		todolist = modifyItem(todolist, todonow, "", 0, ACTION_MOD_PRIORITY)
		finish(todolist)
		return
	}

	// set item to important
	if todoimp > -1 {
		todolist = modifyItem(todolist, todoimp, "", 1, ACTION_MOD_PRIORITY)
		finish(todolist)
		return
	}

	// set item to normal
	if normal > -1 {
		todolist = modifyItem(todolist, normal, "", 2, ACTION_MOD_PRIORITY)
		finish(todolist)
		return
	}

	// save the todo file in a new location
	if newTodoFilePath != "" {
		// check for trainling slash - crude way but works
		if !strings.HasSuffix(newTodoFilePath, "/") {
			newTodoFilePath = newTodoFilePath + "/"
		}
		// go ser it up
		setupRemote(newTodoFilePath)
		return
	}

	// add or delete an item to/from todo list
	if len(arguments) > 0 {
		// check if a number was passed
		n, err := strconv.Atoi(flag.Arg(0))
		//if yes, then remove that item number from the todo list
		if err == nil {
			// remove item
			todolist = removeItem(todolist, n)
		} else {
			// check if the new item has ** and add it as -p 1
			// or TODAY and add it as a -p 0
			if strings.Index(flag.Arg(0), "**") > -1 {
				// Remove the "** " from the string and add the rest
				newitemstring := strings.ReplaceAll(flag.Arg(0), "** ", "")
				// add item
				todolist = addItem(todolist, newitemstring, 1)
			} else if strings.Index(flag.Arg(0), "TODAY") > -1 {
				// Remove the "TODAY " from the string and add the rest
				newitemstring := strings.ReplaceAll(flag.Arg(0), "TODAY ", "")
				// add item
				todolist = addItem(todolist, newitemstring, 0)
			} else {
				// add item
				todolist = addItem(todolist, flag.Arg(0), priority)
			}
		}
	}

	finish(todolist)
}

func init() {
	// setup empty todo file if it doesn't exist
	setupLocal()
	readConfigFile(configfile)
}
